#include "solaris.h"

int uif_errno = 0;

/* 
 * Solaris 7 implementation of libifconfig. This stuff is 
 * all based on SIOCGIFCONF/SIOCSIFADDR/etc. 
 */

static int acquire_socket(int family);
static int get_ifs(struct md_block *mdb);
static int init_ifr(struct md_block *mdb, char *dev, int family);
static int fill_mdb(struct md_block *mdb);
static int set_ifip4brd(struct md_block *mdb, u_int32_t ip);
static int set_ifip4addr(struct md_block *mdb, u_int32_t ip);
static int set_ifip4dstaddr(struct md_block *mdb, u_int32_t ip);
static int set_ifip4mask(struct md_block *mdb, u_int32_t ip);
static int set_ifmdflags(struct md_block *mdb, u_int32_t flags);


/* ---------------------------------------------------------------- */

void *
uif_get(char *dev, int family) {
	struct md_block *mdb = xcalloc(1, sizeof(*mdb));

	assert(dev);

	if(!family)
		mdb->family = AF_INET;
	else
		mdb->family = family;

	mdb->fd = -1;

	if(init_ifr(mdb, dev, family) < 0) 
		return(NULL);

	return((void *) mdb);	
}

/* ---------------------------------------------------------------- */

int
uif_list(int family, char ***names) {
	struct md_block mdb;
	char **np = NULL;
	int i = 0;

	if(!family)
		family = AF_INET;

	assert(names);

	memset(&mdb, 0, sizeof(mdb));

	mdb.fd = acquire_socket(family);
	if(mdb.fd < 0) {
		uif_errno = UIFE_OPEN;
		return(-1);
	}

	if(get_ifs(&mdb) < 0) {
		uif_errno = UIFE_IFLIST;
		if(mdb.aliases)
			FREE(mdb.aliases);

		return(-1);
	}

	if(!mdb.aliases)
		return(0);

	np = xcalloc(mdb.aliascnt, sizeof(char *));
	
	for(i = 0; i < mdb.aliascnt; i++) 
		np[i] = xstrdup(mdb.aliases[i].ifr_name);

	FREE(mdb.aliases);

	*names = np;

	return(i);	
}

/* ---------------------------------------------------------------- */

u_int32_t 
uif_getflags(void *handle, int *err) {
	struct md_block *mdb = (struct md_block *) handle;
	u_int32_t flagword = 0;

	if(err)
		*err = 0;

	assert(handle);

	if(fill_mdb(mdb) < 0) {
		if(err) 
			*err = -1;
		return(0);
	}

	if(mdb->flags & IFF_UP)
		flagword |= UIF_UP;

	if(mdb->flags & IFF_LOOPBACK)
		flagword |= UIF_LOOP;

	if(!(mdb->flags & IFF_NOARP))
		flagword |= UIF_ARP;

	if(mdb->flags & IFF_PROMISC)
		flagword |= UIF_PROMISC;

	if(!(mdb->flags & IFF_NOTRAILERS))
		flagword |= UIF_TRAILERS;

	if(mdb->flags & IFF_ALLMULTI)
		flagword |= UIF_ALLMULTI;

	if(mdb->flags & IFF_MULTICAST)
		flagword |= UIF_ANYMULTI;

	if(mdb->flags & IFF_DYNAMIC)
		flagword |= UIF_DYNAMIC;

	return(flagword);
}

/* ---------------------------------------------------------------- */

int
uif_setflags(void *handle, u_int32_t flagword) {
	struct md_block *mdb = (struct md_block *) handle;

	short word = 0;

	assert(handle);

	if(flagword & UIF_UP)
		word |= (IFF_UP|IFF_RUNNING);
	
	if(flagword & UIF_LOOP)
		word |= IFF_LOOPBACK;

	if(!(flagword & UIF_ARP))
		word |= IFF_NOARP;

	if(flagword & UIF_PROMISC) 
		word |= IFF_PROMISC;
	
	if(!(flagword & UIF_TRAILERS))
		word |= IFF_NOTRAILERS;

	/* If we enable multicast descrimination, we should
	 * have multicast enabled.
	 */

	if(flagword & UIF_ALLMULTI) 
		word |= (IFF_MULTICAST|IFF_ALLMULTI);

	if(flagword & UIF_ANYMULTI)
		word |= IFF_MULTICAST;

	if(flagword & UIF_DYNAMIC)
		word |= IFF_DYNAMIC;

	return(set_ifmdflags(mdb, word));
}

/* ---------------------------------------------------------------- */

void 
uif_release(void *handle) {
	struct md_block *mdb = (struct md_block *) handle;

	assert(handle);

	close(mdb->fd);

	if(mdb->aliases)
		FREE(mdb->aliases);

	FREE(mdb);

	return;
}

/* ---------------------------------------------------------------- */

int
uif_setmdflags(void *handle, u_int32_t mdflags) {
	struct md_block *mdb = (struct md_block *) handle;

	assert(handle);

	return(set_ifmdflags(mdb, mdflags));
}

/* ---------------------------------------------------------------- */

u_int32_t 
uif_getip4addr(void *handle, int *err) {
	struct md_block *mdb = (struct md_block *) handle;

	assert(handle);

	if(err) 
		*err = 0;

	if(fill_mdb(mdb) < 0) {
		if(err) 
			*err = -1;
		return(0);
	}

	return(mdb->pa.sin_addr.s_addr);
}

/* ---------------------------------------------------------------- */

int 
uif_setip4addr(void *handle, u_int32_t addr) {
	struct md_block *mdb = (struct md_block *) handle;

	assert(handle);

	if(set_ifip4addr(mdb, addr) < 0) {
		return(-1);
	}

	if(mdb->raw) 
		return(0);

	/* Default the interface to 255.255.255.0 */

	if(uif_setip4mask(mdb, 0xFFFFFF00L) < 0) {
		/* Best effort */
	}

	return(0);
}

/* ---------------------------------------------------------------- */

u_int32_t 
uif_getip4mask(void *handle, int *err) {
	struct md_block *mdb = (struct md_block *) handle;

	assert(handle);

	if(err) 
		*err = 0;

	if(fill_mdb(mdb) < 0) {
		if(err) 
			*err = -1;
		return(0);
	}

	return(mdb->nm.sin_addr.s_addr);
}

/* ---------------------------------------------------------------- */

int 
uif_setip4mask(void *handle, u_int32_t addr) {
	struct md_block *mdb = (struct md_block *) handle;

	assert(handle);

	if(set_ifip4mask(mdb, addr) < 0) {
		return(-1);
	}

	if(!mdb->raw) 
		return(0);

	/* Default the broadcast address to the network's all-1's 
	 * address. We always want to do this.
 	 */

	if(mdb->pa.sin_addr.s_addr) {
		if(uif_setip4bcst(mdb, mdb->pa.sin_addr.s_addr | ~addr) < 0) {
			/* Best effort */
		}
	}

	return(0);
}

/* ---------------------------------------------------------------- */

u_int32_t 
uif_getip4bcst(void *handle, int *err) {
	struct md_block *mdb = (struct md_block *) handle;

	assert(handle);

	if(err) 
		*err = 0;

	if(fill_mdb(mdb) < 0) {
		if(err) 
			*err = -1;
		return(0);
	}

	return(mdb->ba.sin_addr.s_addr);
}

/* ---------------------------------------------------------------- */

int 
uif_setip4bcst(void *handle, u_int32_t addr) {
	struct md_block *mdb = (struct md_block *) handle;

	assert(handle);

	return(set_ifip4brd(mdb, addr));
}

/* ---------------------------------------------------------------- */

u_int32_t 
uif_getip4dst(void *handle, int *err) {
	struct md_block *mdb = (struct md_block *) handle;

	assert(handle);

	if(err) 
		*err = 0;

	if(fill_mdb(mdb) < 0) {
		if(err) 
			*err = -1;
		return(0);
	}

	return(mdb->da.sin_addr.s_addr);
}

/* ---------------------------------------------------------------- */

int 
uif_setip4dst(void *handle, u_int32_t addr) {
	struct md_block *mdb = (struct md_block *) handle;

	assert(handle);

	return(set_ifip4dstaddr(mdb, addr));
}

/* ---------------------------------------------------------------- */

void uif_setraw(void *handle, int v) { 
	struct md_block *mdb = (struct md_block *) handle;

	assert(handle);

	mdb->raw = v;

	return;
}

/* ---------------------------------------------------------------- */

int uif_getraw(void *handle) {
	struct md_block *mdb = (struct md_block *) handle;

	assert(handle);

	return(mdb->raw);
}

/* ---------------------------------------------------------------- */

/* 
 * Linux CONFIG_IP_ALIAS interface alias management.
 *
 * This is painful but not hard to grok. Each alias has
 * an interface entry (an ifreq). Aliases are associated
 * with a primary device by their name and the magic
 * trailer ":%d", where %d is the alias number. To create
 * a new alias, you come up with a new alias number
 * and call SIOCSIFADDR on that ifreq name to assign it
 * an address. Poof, you have an alias. 
 */

u_int32_t *uif_getip4aliases(void *handle, int *count) {
	struct md_block *mdb = (struct md_block *) handle;
	u_int32_t *buf = NULL;
	int i = 0;

	assert(handle);

	/* Always refresh the alias list; we do this in all the
	 * alias functions, and we have no choice --- the list 
	 * could change between functions and we depend on it. 
 	 */

	if(get_ifs(mdb) < 0) {
		uif_errno = UIFE_IFLIST;

		*count = -1;
		return(NULL);
	}

	if(mdb->aliascnt == 0) {
		*count = 0;
		return(NULL);
	}

	/* Extract the addresses from the ifreq list. */

	buf = xcalloc(mdb->aliascnt, sizeof(*buf));

	for(i = 0; i < mdb->aliascnt; i++) {
		struct sockaddr_in *sip = (struct sockaddr_in *)
				&mdb->aliases[i].ifr_addr;

		buf[i] = sip->sin_addr.s_addr;
	}

	*count = i;
	return(buf);
}

/* ---------------------------------------------------------------- */

int 
uif_addip4alias(void *handle, u_int32_t ip) {
	struct md_block *mdb = (struct md_block *) handle;
	int i = 0;
	int maxv = 0;

	assert(handle);

	if(get_ifs(mdb) < 0) {
		uif_errno = UIFE_IFLIST;
		return(-1);
	}

	/* This is detestable and nasty and I blame Linux. Yeah,
	 * I said it. We have to go through this silly interface
	 * list and actually convert strings (strings! drawings
	 * of numbers!) to figure out what the first unused alias
	 * address is.
	 */

	for(i = 0; i < mdb->aliascnt; i++) {
		char *cp = NULL;

		if((cp = strchr(mdb->aliases[i].ifr_name, ':'))) {
			char *cn = cp;

			*cp++ = 0;

			if(!strcmp(mdb->aliases[i].ifr_name, mdb->ifr.ifr_name)) {
				int v = strtol(cp, NULL, 10);

				if(v > maxv)
					maxv = v;
			}

			*cn = ':';
		}		
	}

	{ 
		struct ifreq nifr;

		/* New interface is "baseX:Y", ie, "eth0:1" for alias 1 */

		snprintf(nifr.ifr_name, IFNAMSIZ - 1, "%s:%d", 
			mdb->ifr.ifr_name, maxv + 1);
		nifr.ifr_name[IFNAMSIZ - 1] = 0;

		/* Give it an address. It lives! */

		(*(struct sockaddr_in *)&nifr.ifr_addr).sin_family = AF_INET;
		(*(struct sockaddr_in *)&nifr.ifr_addr).sin_addr.s_addr = ip;

		if(ioctl(mdb->fd, SIOCSIFADDR, &nifr) < 0) {
			uif_errno = UIFE_SETADDR;
			return(-1);
		}	

		/* Copy the flags from the base interface */

		nifr.ifr_flags = mdb->flags;

		if(ioctl(mdb->fd, SIOCSIFFLAGS, &nifr) < 0) {
			uif_errno = UIFE_SETFLAGS;
			return(-1);
		}
	}

	return(0);
}

/* ---------------------------------------------------------------- */

int 
uif_delip4alias(void *handle, u_int32_t ip) {
	struct md_block *mdb = (struct md_block *) handle;
	int i = 0;

	assert(handle);

	if(get_ifs(mdb) < 0) {
		uif_errno = UIFE_IFLIST;
		return(-1);
	}

	/* This is detestable and nasty and I blame Linux. Yeah,
	 * I said it. 
	 */

	for(i = 0; i < mdb->aliascnt; i++) {
		if(strchr(mdb->aliases[i].ifr_name, ':')) {
			if((*(struct sockaddr_in *)&mdb->aliases[i].ifr_addr).	
				sin_addr.s_addr == ip) {

				struct ifreq nifr;

				strcpy(nifr.ifr_name, mdb->aliases[i].ifr_name);

				nifr.ifr_flags = 0; /* DOWN */
				
				if(ioctl(mdb->fd, SIOCSIFFLAGS, &nifr) < 0) {
					uif_errno = UIFE_SETFLAGS;
					return(-1);
				}

				/* Remove its address, it dies! */

				(*(struct sockaddr_in *)&nifr.ifr_addr)
					.sin_family = AF_INET;
				(*(struct sockaddr_in *)&nifr.ifr_addr)
					.sin_addr.s_addr = 0;

				if(ioctl(mdb->fd, SIOCSIFADDR, &nifr) < 0) {
					uif_errno = UIFE_SETADDR;
					return(-1);
				}	

				break;
			}
		}
	}
	
	if(i == mdb->aliascnt)
		return(0);

	return(1);
}

/* ---------------------------------------------------------------- */

static int
acquire_socket(int family) {
	int sfd = socket(family, SOCK_DGRAM, 0);
	return(sfd);
}

/* ---------------------------------------------------------------- */

/* Setting flags == SIOCSIFFLAGS */

static int
set_ifmdflags(struct md_block *mdb, u_int32_t flags) {
	mdb->ifr.ifr_flags = (short) flags;

	if(ioctl(mdb->fd, SIOCSIFFLAGS, &mdb->ifr) < 0) {
		uif_errno = UIFE_SETFLAGS;
		return(-1);
	}

	mdb->flags = flags;

	return(0);
}

/* ---------------------------------------------------------------- */

/* Setting primary address = SIOCSIFADDR */

static int
set_ifip4addr(struct md_block *mdb, u_int32_t ip) {

	((struct sockaddr_in *)&mdb->ifr.ifr_addr)->sin_addr.s_addr = ip;

	if(ioctl(mdb->fd, SIOCSIFADDR, &mdb->ifr) < 0) {
		uif_errno = UIFE_SETADDR;
		return(-1);
	}

	mdb->pa.sin_addr.s_addr = ip;

	return(0);
}

/* ---------------------------------------------------------------- */

/* Setting Destination Address = SIOCSIFDSTADDR */

static int
set_ifip4dstaddr(struct md_block *mdb, u_int32_t ip) {

	((struct sockaddr_in *)&mdb->ifr.ifr_dstaddr)->sin_addr.s_addr = ip;

	if(ioctl(mdb->fd, SIOCSIFDSTADDR, &mdb->ifr) < 0) {
		uif_errno = UIFE_SETDSTADDR;
		return(-1);
	}

	mdb->da.sin_addr.s_addr = ip;

	return(0);
}

/* ---------------------------------------------------------------- */

/* Setting Netmask = SIOCSIFNETMASK */

static int
set_ifip4mask(struct md_block *mdb, u_int32_t ip) {

	((struct sockaddr_in *)&mdb->ifr.ifr_addr)->sin_addr.s_addr = ip;

	if(ioctl(mdb->fd, SIOCSIFNETMASK, &mdb->ifr) < 0) {
		uif_errno = UIFE_SETNETMASK;
		return(-1);
	}

	mdb->nm.sin_addr.s_addr = ip;

	return(0);
}

/* ---------------------------------------------------------------- */

/* Setting Broadcast Address = SIOCSIFBRDADDR */

static int
set_ifip4brd(struct md_block *mdb, u_int32_t ip) {

	((struct sockaddr_in *)&mdb->ifr.ifr_broadaddr)->sin_addr.s_addr = ip;

	if(ioctl(mdb->fd, SIOCSIFBRDADDR, &mdb->ifr) < 0) {
		uif_errno = UIFE_SETBCAST;
		return(-1);
	}

	mdb->ba.sin_addr.s_addr = ip;

	return(0);
}

/* ---------------------------------------------------------------- */

/* Get as much information about the interface as we can and stash
 * it away. 
 */

static int
fill_mdb(struct md_block *mdb) {
	int sfd = mdb->fd;

	mdb->aliascnt = 0;
	mdb->aliases  = NULL;

	if(ioctl(sfd, SIOCGIFADDR, &mdb->ifr) < 0)  {
		uif_errno = UIFE_GETADDR;
		return(-1);
	}

	memcpy(&mdb->pa, &mdb->ifr.ifr_addr, sizeof(mdb->pa));

	if(ioctl(sfd, SIOCGIFFLAGS, &mdb->ifr) < 0) {
		uif_errno = UIFE_GETFLAGS;
		return(-1);
	}

	mdb->flags = mdb->ifr.ifr_flags;

	if(ioctl(sfd, SIOCGIFDSTADDR, &mdb->ifr) < 0) {
		uif_errno = UIFE_GETDSTADDR;
		mdb->da.sin_addr.s_addr = INADDR_NONE;
	} else 
		memcpy(&mdb->da, &mdb->ifr.ifr_dstaddr, sizeof(mdb->da));

	if(ioctl(sfd, SIOCGIFBRDADDR, &mdb->ifr) < 0) {
		uif_errno = UIFE_GETBCAST;
		mdb->da.sin_addr.s_addr = 0;
	} else 
		memcpy(&mdb->ba, &mdb->ifr.ifr_broadaddr, sizeof(mdb->ba));

	if(ioctl(sfd, SIOCGIFNETMASK, &mdb->ifr) < 0) {
		uif_errno = UIFE_GETNETMASK;
		return(-1);
	}

	memcpy(&mdb->nm, &mdb->ifr.ifr_addr, sizeof(mdb->nm));

	return(0);
}

/* ---------------------------------------------------------------- */

static int
init_ifr(struct md_block *mdb, char *dev, int family) {

	/* Apply semantics to address sets */

	mdb->raw = 0;

	mdb->flags = (IFF_UP|IFF_RUNNING);

	memset(&mdb->ifr, 0, sizeof(struct ifreq));

	strncpy(mdb->ifr.ifr_name, dev, IFNAMSIZ);

	if(mdb->fd == -1) {
		mdb->fd = acquire_socket(mdb->family);
		if(mdb->fd < 0) {
			uif_errno = UIFE_OPEN;
			return(-1);	
		}
	}

	if(fill_mdb(mdb) < 0)
		return(-1);

	return(0);
}

/* ---------------------------------------------------------------- */

static int
get_ifs(struct md_block *mdb) {
	struct ifconf ic;
	int sz;

	if(mdb->aliases) {
		FREE(mdb->aliases);
		mdb->aliascnt = 0;
		mdb->aliases = NULL;
	}

	memset(&ic, 0, sizeof(ic));

	sz = 512;
	ic.ifc_buf = NULL;

	do {
		sz *= 2;

		if(ic.ifc_buf)
			FREE(ic.ifc_buf);

		ic.ifc_len = sz;
		ic.ifc_buf = xalloc(sz);

		mdb->aliases = (struct ifreq *) ic.ifc_buf;

		if(ioctl(mdb->fd, SIOCGIFCONF, &ic) < 0) {
			return(-1);
		}
	} while(ic.ifc_len * sizeof(*mdb->aliases) > sz);

	mdb->aliascnt = ic.ifc_len;	

	/* XXX
	 * Something is wrong here; do I really need to do 
	 * this? What are these wierdo blank interfaces doing
	 * here?
	 */

	{ 
		int inc = 0;

		for(inc = 0; inc < mdb->aliascnt; inc++) 
			if(!mdb->aliases[inc].ifr_name[0])
				break;
				
		mdb->aliascnt = inc;	
	} 

	return(0);
}

/* ---------------------------------------------------------------- */



